// RELATED SOFTWARE: Scilab >= 5.5.0
// PACKAGE  : uman : Unified user MANual for Scilab and external modules
// VERSION  : 1.4 - 2015-07-31
// AUTHOR   : Samuel GOUGEON - private individual
// COPYRIGHT: Samuel GOUGEON - 2015
// LICENSE for uman : CeCILL : http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
// 
// DISTRIBUTION: http://atoms.scilab.org/toolboxes/uman
//
// DISCLAIMER: This toolbox is provided as is, with NO WARRANTY of any kind.
//   Use it to your own risks & responsability.

function uman(varargin)
    global %helps

    // FORMATING and partially PARSING OPTIONS
    // ---------------------------------------
    if argn(2)==0 then
        uman("uman","e")
        return
    end
    if argn(2)==3 then
        tmp = convstr(varargin(3))
        if grep(tmp,"l")~=[]
            varargin(1) = varargin(1)+tmp
        else
            varargin(1) = "-"+tmp+part(varargin(1), 2:$)
        end
    end
    if argn(2)==1 then
        if varargin(1)=="-"  then
            fun0 = "-"
            options = ""
        elseif part(varargin(1),1)=="-"
            if grep(varargin(1),"c")~=[] then
                clc
            end
            uman("uman")
            return
        else
            fun0 = varargin(1)
            options = ""
        end
    else
        if varargin(1)=="-" | part(varargin(1),1)~="-" then
            fun0 = varargin(1)
            options = varargin(2)
        else
            fun0 = varargin(2)
            options = varargin(1)
        end
    end
    
    // OPTIONS ANALYSIS
    // ----------------
    options0 = convstr(options)  // useful in case of redirection (recursive => call)
    i = strindex(options0, "l")
    if i~=[] then
        LANG    = part(options0, (i(1)+1):(i(1)+2))
        if LANG==""
            LANG = "en"
        end
        options = uman_strsubst(options, ["/l../i" "/l$/i"], ["" ""], "r")
    else
        LANG = part(getlanguage(), 1:2)
    end
    
    fun = fun0
    if grep(options, "a")~=[] then
        options = strsubst(options,"a","deh")
    end
    if grep(options, "x")~=[] then
        fun  = redirect_from_external(fun, options)
        // Some equivalences have "£" inserted: prefixing chars are options to force
        //  By default, "s" is forced, meaning that the equivalence is designed
        //  to target a parent summary:
        if grep(fun,"£")~=[] then
            s = strtok(fun, "£")
            fun = part(strchr(fun,"£"), 2:$)
            if s==fun
                if grep(options, "s")==[]
                    options = "s" + options
                end
            else
                options = s + options
            end
        end
    end
    fun = uman_strsubst(fun,"/^%/" ,"percent","r") 
    // This substitution will have to be cancelled in "wb" mode before calling
    //  the bug page. Nevertheless, it is required to search the right page
    //  to determine the related module (or scilab) and the related bug page 
    //  to be opened.
    
    // option "s" (summary) kills option "b" (bug) and "w" (web)
    if grep(options, "s")~=[] then
        options = uman_strsubst(options, ["b" "w"], ["" ""])
    end
    bugOnWeb = grep(options, "w")~=[] & grep(options,"b")~=[]

    // SETTING THE LANGUAGE
    // --------------------
    supportedLANG = [ "en_US" "fr_FR" "ja_JP" "de_DE" "pt_BR" "ru_RU" "zh_CN"]
    i = grep(supportedLANG, "/^"+LANG+"_/", "r")
    if i==[] then
        LANG = "en_US"
    else
        LANG = supportedLANG(i)
    end

    // REDIRECTIONS for Symbols. Translation for "%" in fun names
    // ----------------------------------------------------------
    if ~bugOnWeb then
        fun = uman_strsubst(fun,["/\{/" "/\}/"],["[" "]"],"r") 
        sub = "assignation,great,  > ,  < ,   [    ,   ]    ,  :  ,   //   ,"
        ids = "insertion  ,less ,less,less,brackets,brackets,colon,comments,"
        
        sub = sub+"    <>    ,   ~=     ,   <=     ,   >=     ,"
        ids = ids+"comparison,comparison,comparison,comparison,"

        sub = sub+"  =  ,    ==    ,"
        ids = ids+"equal,comparison,"
        // bug with these ones, due to Scilab parser. 
        // Must be set within quotes: uman "=="
        
        sub = sub+"  &   ,  |  , ~ ,   $  , . , []  ,"
        ids = ids+"and_op,or_op,not,dollar,dot,empty,"
            // % automaticaly replaced with percent 
        
        sub = sub+"  -  ,  + ,  * , .* ,  /  , ./  ,  /. , .^  , **  , ^ , .*.,    \    ,"
        ids = ids+"minus,plus,star,star,slash,slash,slash,power,power,hat,kron,backslash,"
        
        sub = sub+"    \.     ,     *.    ,    ./.    ,    .\.    ,"
        ids = ids+"overloading,overloading,overloading,overloading,"
        
        sub = sub+"  '' ,    ;    "
        ids = ids+"quote,semicolon"
        
        sub = sub+""
        ids = ids+""
        sub = [ stripblanks(strsplit(sub, ",")) ; "," ]
        ids = [ stripblanks(strsplit(ids, ",")) ; "comma" ]
        i = find(fun==sub)
        if i~=[] then
            fun = ids(i(1))
        end
        if grep(fun, " ") then
            fun = strsubst(fun, " ", "_")
        end
    end

    // If the query includes "property" or "properties", "d" description flag 
    //  is made mandatory
    if grep(convstr(fun),"propert")~=[] & grep(options, "d")==[] then
        options = "d" + options
    end

    // OTHER INITIALISATIONS
    // ---------------------
    // Case sensitivity
    if convstr(fun)==fun then
        casse = "-C"
    else
        casse = ""
    end
    // unzip application vs OS
    if getos() == "Windows" then
        unzip  = pathconvert(SCI+"\tools\zip\") + "unzip.exe "+casse+" -o -q " 
        //unzip2 = pathconvert(SCI+"\tools\gzip\") + "gzip.exe "
    else
        unzip = "unzip "+casse+" "
    end

    fs = filesep()
    global uman_alreadyWebed uman_HeadCommented
    uman_alreadyWebed = %f
    uman_HeadCommented = %f    
    htmlFound = %f
    unzippedFile = ""        // = flag for "File Not Found"

    manDir = pathconvert(SCIHOME + "\uman\")   // storage of unzipped HTML files
    if fileinfo(manDir)==[] then  
        mkdir(manDir)
    end
    manAtomsDir = pathconvert(SCIHOME + "\uman\atoms\")  // storage of unzipped HTML files
    if fileinfo(manAtomsDir)==[] then  
        mkdir(manAtomsDir)
    else
        // Refresh option => deleting all extracted files in manAtomsDir
        if grep(options, "r")~=[] then
            rmdir(manAtomsDir,"s")
            mkdir(manAtomsDir)
        end
    end
    
    mapDir = manDir + "jhelpMaps" + fs          // storage of jhelpmap.jhm files
    if grep(options, "r")~=[] then
        rmdir(mapDir,"s")
    end
    if ~isdir(mapDir) then
        mkdir(mapDir)
    end

    subDirEN   = "scilab_en_US_help"
    subDirLANG = "scilab_" + LANG + "_help"

    atoms = atomsGetInstalled()
    if atoms~=[] then
        atoms(:,4) = getshortpathname(pathconvert(atoms(:,4)))    
    end
    module = ""
    
    // SEARCHING FOR THE RIGHT HELP FILE. MANAGING UNAVAILABLE ONE. ZIP EXTRACTION
    // ===========================================================================
    // We need this part also for bugOnWeb: if the pattern matches an external
    //  function (e.g. from ATOMS), the external module must be identified
    //  in order to target the right online bug report place
    jarListFile = manDir+"jarListe.txt"
    i = fileinfo(jarListFile)
    // If the jar list does not yet exist, or if it is more than 7 days old,
    //  or if the "r" (reload / refresh) option is used, we rebuild and record 
    //  it and re-extract jhelpmap.jhm files :
    if i==[] | (getdate("s")-i(6))>7*24*3600 | grep(options,"r")~=[] then
        jarList = uman_listJar()
        csvWrite(jarList, jarListFile, ascii(9))
        extractJhelpMaps(jarList, atoms, unzip, manDir, mapDir)
    else
        jarList = mgetl(jarListFile)
    end
    
    // Looking for the pattern in jhelpmap files 
    //  => related jar & html filename
    // -----------------------------------------
    // Sorting jhm files to be scanned
    jhmTab = listfiles(mapDir)     // jhmTab(1) == "scilab.jhm"
    i = grep(jhmTab, "scilab.jhm")
    jhmTab(i) = []
    if grep(options, "x")==[] then
        jhmTab = ["scilab.jhm" ; jhmTab]
    else
        jhmTab($+1) = "scilab.jhm"
    end
    // scanning jhm files
    htmlName = ""

    if casse=="-C"
        grepCase = "i"
    else
        grepCase = ""
    end
    
    if (~bugOnWeb | grep(fun,"|")==[]) & grep(fun, ["/" "\" "*" ":"])==[]
        for jhm = jhmTab'
            //doc = xmlRead(jhm)
            //e = xmlXPath(doc,"//mapID[@target="""+fun+"""]")
           doc = mgetl(mapDir+jhm)
           i = grep(doc, "/target="""+fun+"""/"+grepCase, "r")
           if length(i)>1
               i = grep(doc, "/target="""+fun+"""/", "r")
           end
           if i~=[]
               search = "/^.*?target="""+fun+""" url=""(.+?.html).*?"".*$/"+grepCase
               htmlName = uman_strsubst(doc(i), search, "$1","r")
               module = strtok(jhm,".")
               break
           end
        end
    end
    inATOMS = (find(atoms(:,1)==module)~=[])

   // item found in the doc
   // 

   if htmlName~="" then
       // in which jar file ?
       if module=="scilab"
           epath = pathconvert("\modules\helptools\jar\scilab_££_help.jar",%f)
       else
           epath = fs + module + fs + ".*?jar"+fs+"scilab_££_help.jar"
       end
       epath = strsubst(epath, "\", "\\")
       i = grep(jarList, "|^.*?"+strsubst(epath,"££",LANG)+"|i", "r")
       if i==[]
           i = grep(jarList, "|"+strsubst(epath, "££", "en_US")+"|i", "r")
           subDir = subDirEN
       else
           subDir = subDirLANG
       end
       jar = jarList(i(1))

       // is the HTML of this jar unzipped ?
       zippedFile = subDir + fs + htmlName
        if module=="scilab" 
            outDir = manDir         // a un / final
        else  // ATOMS and other contribs are mixed
            outDir = manAtomsDir    // idem
        end
        unzipped = outDir + subDir + fs + htmlName
       if isfile(unzipped) & grep(options, "r")==[]
           unzippedFile = unzipped
       else
           if isfile(unzipped) 
               deletefile(unzipped)
           end
           // We extract the HTML file:
           command = unzip + jar + " " + zippedFile + " -d " + outDir
           status = host(command)
           if status==0 // file found and extracted
               unzippedFile = unzipped
           end
       end
    else // item not found in the doc
         // =========================
        if grep(options, "x")==[] & ~isdef("uman_already_selfcalled", "n") //&..
//            (grep(options, "w")==[] | grep(options, "b")==[])
            fun  = redirect_from_external(fun, options)
            // Some equivalences have "£" inserted: prefixing chars are options to force
            //  By default, "s" is forced, meaning that the equivalence is designed
            //  to target a parent summary:
            if grep(fun,"£")~=[] then
                s = strtok(fun, "£")
                fun = part(strchr(fun,"£"), 2:$)
                if s==fun
                    if grep(options, "s")==[]
                        options = "s" + options
                    end
                else
                    options = s + options
                end
            end
            // option "s" (summary) kills option "b" (bug)
            //if grep(options, "s")~=[] then  
            //    options = strsubst(options, "b", "")
            //end   // This cancellation is actually quite useless
                    // It kills the mixed "wb" feature. Rather bad.
            bugOnWeb = grep(options, "w")~=[] & grep(options,"b")~=[]
            if ~bugOnWeb 
                uman_already_selfcalled = %t
                uman(fun, options+"L"+LANG)
            end
        end
    end

    // bugs report places
    // ------------------
    // default place:
    BugzillaSearch = "http://bugzilla.scilab.org/buglist.cgi"+..
      "?query_format=advanced"+..
      "&o1=regexp&f1=short_desc&v1=(^|[^a-zA-Z])£([^0-9a-zA-Z_%3D%3B]|$)"+..
      "&bug_status=NEW&bug_status=ASSIGNED&bug_status=REOPENED"+..
      "&bug_status=RESOLVED&short_desc_type=allwordssubstr"+..
      "&order=resolution,bug_id DESC"
      // £ will be replaced with the item name
        
    // for registered ATOMS modules
    AF = get_atoms_forge()
    if module~="" then
        k = find(AF(:,1)==module)
    else
        k = find(AF(:,1)==fun)
    end
    isAnAtoms = find(atoms(:,1)==module)~=[]
    if module~="scilab" & k~=[] then
        forgeName = AF(k,2)
        if forgeName~=""
            BugReport = "http://forge.scilab.org/index.php/p/"+forgeName+"/issues/"
            BugzillaSearch = BugReport + "search/?q="
            if unzippedFile~=""
                BugzillaSearch = BugzillaSearch + fun
            else
                BugzillaSearch = BugzillaSearch + fun0
            end
        else
            if isAnAtoms
                tmp = "http://atoms.scilab.org/toolboxes/"+module+"/#comments_remove"
            else
                tmp = stripblanks(AF(k,3))
            end
            BugReport = tmp
            BugzillaSearch = tmp
        end
        
    // ATOMS modules unregistered by uman:
    elseif find(atoms(:,1)==module)~=[]
        website = "http://atoms.scilab.org/toolboxes/"+module+"/#comments_remove"
        BugReport = website
        BugzillaSearch = website
    
    // for Contrib modules 
    elseif module~="scilab" & htmlName~=""
        BugReport = ""
        BugzillaSearch = ""
    end

    // ---------------------------------------------------------------

    // DISPLAY
    // =======
    msg1 = ""
    // Bugzilla | ticket/forge | website
    // ---------------------------------
    if bugOnWeb & BugzillaSearch~=""  then
        fun = strsubst(fun, "/^percent/", "%", "r")
        //symbols = "><[]:<>~=<=>==&|$.**-+././.^.*.\.'';"
        if grep(fun,"/")~=[]
            delim = "!"
        else
            delim = "/"
        end
        fun = uman_strsubst(fun, ..
            delim+[ "&"   "\*"  "\."  "\-"  "\:"  "\]"  "\["  "\("  "\)"  "\$"    "\^" "/" ]+delim, ..
                  [ "%26" "\\*" "\\." "\\-" "\\:" "\\]" "\\[" "\\(" "\\)" "\\%24" "\\^"  "%2F"  ], "r")
        url = strsubst(BugzillaSearch,"£",fun0)
        msg = _("uman","%s: ""w"" option is not supported under ""%s""\n")
        if ~isdef("uman_already_selfcalled","n")
            uman_openURL(url, msg)
            if grep(options, ["s" "g"])==[]
                return
            end
        end
        return
    end

    // targetting ONLINE help page (or site)
    // ....................................
    printWeb = ( (  ( grep(options,"w")~=[]  &  grep(options,"s")==[] ) | ..
                    ( unzippedFile=="" & grep(options0,"w")~=[] ) ..
                 ) &  or(module==["" "scilab"]) & grep(options,"b")==[]) 
                 
    if printWeb & ~uman_alreadyWebed then
        url = "http://help.scilab.org/docs/current/"
        supported_online= ["en_US" "fr_FR" "pt_BR" "ru_RU" "ja_JP"]
        if and(LANG~=supported_online)
            langH = "en_US"
        else
            langH = LANG
        end
        if unzippedFile~=""
            // url = url+langH+"/"+fun+".html"
            url = url+langH+"/"+basename(unzippedFile)+".html"
        else
            tmp = get_atoms_forge()
            refs = tmp(:,1)
            if convstr(fun0)==fun0
                refs = convstr(refs)
            end
            tmp2 = tmp(refs==fun0, 3)
            if tmp2==""
                tmp2 = tmp(refs==fun0, 1)
                tmp = "http://atoms.scilab.org/toolboxes/"+tmp2
            else
                tmp = tmp2
            end
            // the pattern is a registered entry of man (maybe a FileExchange 
            // one) with a dedicated website
            if tmp~=[]  
                url = strtok(tmp(1),"#")
            else
                url = url + langH + "/search/" + fun0
            end
        end
        msg = _("uman", "%s: ""w"" option is not supported under ""%s""\n")
        msg = msprintf(msg, "uman", getos())
        if url~=""
            uman_openURL(url, msg)
            uman_alreadyWebed = %t
        else
            if unzippedFile=="" & ~uman_alreadyWebed  
                // fun0 may be the name of a module or of an external 
                //   reference, that may have a redirection
                tmp = redirect_from_external(fun0, options)
                if tmp~=fun0
                    uman(tmp, options0)
                end
            end
        end
    end

    // Setting the flag allowing to target the console
    // ...............................................
    //  - whatever is the "g" status (can duplicate the output)
    //  - whether "w" is not used, OR (NOT "wb" AND NOT Scilab module)
    printConsole = (grep(options, "w")==[] | (~bugOnWeb & module~="scilab" ))

    if printConsole & grep(options,"j")==[] 
        diary([], "pause") // http://bugzilla.scilab.org/show_bug.cgi?id=12551
    end

    // The required help file exists...
    // --------------------------------
    if unzippedFile~=""       // we go on on processing
        r = uman_grab(unzippedFile, fun, LANG)
        r.fun = fun0

        // Targetting the console
        // ......................
        msg2 = ""
        if  printConsole
            if grep(options,"c")~=[] then
                clc()
            else    // separating zone
                tmp = lines()
                write(%io(2), part("=", ones(1,tmp(1))))
            end
            // item's page
            // - - - - - -
            msg0 = ""
            if grep(options, "s")==[]
                if module~="scilab"
                    if inATOMS
                        r.bug_report = BugReport
                        r.bugs_known = BugzillaSearch
                    else
                        r.bug_report = ""
                        r.bugs_known = ""
                    end
                end

                uman_disp(r, options)
                
                if grep(options,"w")~=[] & module~="scilab"
                    msg0 = " uman " + fun0 + " " + options0 + " :";
                    msg2 = " ["+module+"] "+..
                           _("uman","module EXTERNAL   => ONLINE help page UNAVAILABLE")
                end
                if msg1~="", msg1 = " [" + module + "] " + msg1, end
                tmp = part("-~", ones(1,max(length([msg1 msg2]))/2).*.[1 2])
                if tmp~=""
                    write(%io(2),tmp)
                    if (msg0~=""), write(%io(2), msg0), end
                    if (msg1~=""), write(%io(2), msg1), end
                    if (msg2~=""), write(%io(2), msg2), end
                    write(%io(2), " "+tmp)
                end
                
            // Parent summary
            // - - - -- - - -
            else 
                tmp = strsplit(unzippedFile, fs)
                zippedFile = tmp($-1) + fs + r.parentSummary + ".html"
                tmp = strsplit(unzippedFile, fs)
                outDir = strcat(tmp(1:$-2), fs)
                unzippedFile = dirname(unzippedFile) + fs + r.parentSummary + ".html"

                if fileinfo(unzippedFile)==[]
                    command = unzip + jar + " " + zippedFile + " -d " + outDir
                    status = host(command)  // status==0  <=> OK
                end
                
                // Parsing the summary page ++++++++++++++++++++++++++
                html = mgetl(unzippedFile)                
                search  = [ "<html>" 
                            "<body>" 
                            "<head>.*?</head>" 
                            "<div class=""manualnavbar"">.*?</div>" 
                            "</body>" 
                            "</html>"
                            "<br />"
                            ]
                search = "|" + search + "|"
                replace = ["<div>" "" "" "" "" "</div>" ascii(10) ]
                html2 = uman_strsubst(strcat(html, " "), search, replace, "r")
                tmp = xmlReadStr(html2)
                simpleContent = %t
                text = uman_xml2txt(tmp.root, " ")
                tmp2 = strsplit(text(1), ">")
                if grep(tmp2(1),"Scilab")~=[] then 
                    tmp2(1) = "Scilab"
                end
                tmp2 = stripblanks(tmp2)
                tmp2(find(tmp2=="")) = []
                if tmp2(1)==tmp2(2)
                    tmp2(2) = []
                end
                text(1) = strcat(tmp2," > ")+ascii(10)
                text(2) = []
                // Removing extra blank lines in deepest sub-blocks
                i = find(text==" ")
                text(i) = []
                // Displaying the summary page +++++++++++++++++++++++
                write(%io(2), text)
                xmlDelete(tmp)
            end
            // targeting the console
        end   

    // The required help file DOES NOT EXIST
    // -------------------------------------
    else                   
        if grep(options, "w")==[]
            // We test if the pattern is an existing function
            ierr = execstr("Fun="+fun0+";","errcatch")
            if (ierr~=0 | typeof(Fun)~="function") & convstr(fun0)==fun0
                // fun0 may be the lower-cased name of another function 
                n = who("local")
                tmp = find(convstr(n)==fun0)
                if tmp~=[]
                    for f = tmp
                        execstr("Fun  = "+n(f))
                        if typeof(Fun)=="function"
                            ierr = 0
                            fun0 = n(f)
                            break
                        end
                    end
                end
            end
            if ierr==0 & typeof(Fun)=="function" then
                if grep(options,"c")~=[] then
                    clc()
                end
                // we display the heading comments (if any)
                if ~isdef("uman_HeadCommented", "n")
                    head_comments(fun0)
                    uman_HeadCommented = %t
                end
            elseif fun~=fun0 & isdef("uman_already_selfcalled","n")
                msg = _("uman %s w  // should work (internet required)\n")
                mprintf(msg,fun)
            elseif isdef("uman_already_selfcalled","n") | grep(options, "x")~=[]
                msg = _("uman","%s: ""%s"" has no dedicated page.\n")
                mprintf(msg,"uman",fun0)
                if find(convstr(AF(:,1))==convstr(fun))~=[]
                    msg = _("uman %s w  // should work (internet required)\n")
                    mprintf(msg,convstr(fun))
                end
            end
        end
    end
    if printConsole & grep(options,"j")==[] 
        diary([], "resume") // http://bugzilla.scilab.org/show_bug.cgi?id=12551
    end
    
    // targetting helpbrowser :
    // ......................
    // must be after print in console, to have called xml_grab() and got r.
    if grep(options,"g")~=[] then 
        if module=="scilab" ..
              | ( inATOMS & atomsIsLoaded(module) ) ..
              | ( ~inATOMS & find(%helps(:,2)==module)~=[] )
            // call the help browser in the chosen language
            if grep(options, "s")~=[]
                tmp = r.parentSummary
            elseif module=="scilab"
                tmp = basename(htmlName)
            else
                tmp = fun
            end
            helpbrowser(%helps(:,1), tmp, LANG, %f)
        elseif module~="" // module known but unloaded
            msg1 = _("uman", "module NOT LOADED => UNVAILABLE in HELP BROWSER")
        elseif isdef("uman_already_selfcalled", "n") // unknown feature
            helpbrowser(%helps(:,1), fun0, LANG, %f)
        end
    end

endfunction

// ===================================================================

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()

function uman_disp(man_contents, options)
    // Finally displays the formatted help text content into the console
    if or(part(LANG,[1 2])==["ja" "zh"])
        lengthFactor = 1.5
    else
        lengthFactor = 1
    end
    u = "----------------------------------------------------------"+..
        "----------------------------------------------------------"
    s = man_contents
    c = %io(2)
    
    t = s.path
    write(c, ["" ; t ; part(".",ones(1,length(t))) ])

    write(c, ["" ; strcat(justify([s.refname s.refpurpose],"l")," - ","c")])
    
    if s.synopsis~=[] then
        t = s.synopsisTitle
        write(c, [ "" ; t ; part(u,1:length(t)*lengthFactor) ; " "+s.synopsis ])
    end
    // Loop over other sections, in the order they appear in the original page
    sections = man_contents.sections
    ns = size(sections,"*")

    for is = 1:ns
        s = sections(is)
        tmp = s.type
        if (or(tmp==["parameters" "see_also"]) | ..
            (grep(options,"d")~=[] & tmp=="description") | ..
            (grep(options,"e")~=[] & tmp=="examples") | ..
            (grep(options,"h")~=[] & tmp=="history"))
            t = s.title
            if t~=""
                write(c, [ "" ; t ; part(u,1:length(t)*lengthFactor)])
            end
            if s.contents~=[]
                if t==""
                    write(c,["";""])
                end
                if tmp=="see_also"
                    s.contents = part(s.contents, 3:$)  // trimming starting stars
                end
                write(c, " "+s.contents)
            end
        end
    end
    // Bugs
    if grep(options,"b")~=[] 
        t = _("uman","Bugs")
        write(c, [ "" ; t ; part(u,1:length(t)*lengthFactor) ])
        write(c, [ _("uman","Known")+"  : " + man_contents.bugs_known ; ..
                   _("uman","Report")+" : " + man_contents.bug_report ])
    end
endfunction

// ===================================================================

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()

function r = uman_grab(htmlf, fun, LANG, options)
    // htmlf : HTML file name
    // fun : function's name or item
    // LANG: ##_## language string
    // options: options string from uman()
    // r. : structure with fields
    //  .fun .title  .path .refname .refpurpose .synopsisTitle .synopsis
    //  .paramTitle  .parameters .examples .seeAlsoTitle .seeAlso 
    //  .bug_report  .bugs_known
    //  .parentSummary

    lang = part(LANG,1:2)
    
    html = mgetl(htmlf)

    search  = ["/<br>/" "/<meta .*?>/" "/<[^a-zA-Z!\/]/" ]
    replace = ["<br/>"     ""             "&lt;"         ]
    html2 = uman_strsubst(strcat(html, " "), search, replace, "r")
    
    // URLs GET's & separators must be replaced with %26 in order to not break
    //  XML parsing. Replacing all cannot be done with TCL uman_strsubst because
    //  "&" has a special meaning in this TCL regsub context 
    // => using a while loop here: 
    src = ""
    while or(html2~=src) do
        src = html2
        html2 = uman_strsubst(src, "|(href=""http:[^\""]*?)&([^\""]*?"")|", "$1%26$2", "r")
    end
    
    doc = xmlReadStr(html2)

    // Table of <pre.. </pre> sections (badly processed as XML)
    [s,e,preTable] = regexp(strcat(html,"££"),"%<pre[\s>].*?</pre>%mi")
    e = xmlXPath(doc,"//pre")
    for i = 1:size(s,"*")
        tmp = preTable(i)
        preTable(i) = uman_cleantxt(preTable(i),"tags|htmlent|style")
        // now we forge an id = full content cleaned
        preTableNoTag(i) = uman_strsubst(e(i).content, [" " "££"], ["" ""]) 
    end

    htmlcat = strcat(html, "££")
    
    r.fun = fun
    
    // Title
    e = xmlXPath(doc,"//title")
    r.title = e.content

    // Path in help
    // ------------
    e = xmlXPath(doc,"//span[@class=""path""]")
    tmp = strsplit(e.content, ">")
    if grep(tmp(1), "Scilab")~=[] then 
        tmp(1) = "Scilab"
    end
    tmp = stripblanks(tmp)
    tmp(find(tmp=="")) = []
    if tmp(1)==tmp(2)
        tmp(2) = []
    end
    r.path = strcat(tmp," > ")
    
    // Parent Summary (from path)
    tmp = "%<span class=""path"">.*?</span>%mi"
    [s, e, tmp] = regexp(html2, tmp)
    tmp2 = "%href="".*?""%mi"
    [s, e, tmp2] = regexp(tmp, tmp2)
    r.parentSummary = part(tmp2($), 7:$-6)

    // URL(report a bug)
    tmp = "%""http://bugzilla\.scilab\.org/enter_bug\.cgi.*?\""%"
    [s,e,m] = regexp(htmlcat, tmp)
    if m~="" then
        r.bug_report = strsubst(m,"""","")+"&short_desc="+fun
    else
        r.bug_report = "http://bugzilla.scilab.org/enter_bug.cgi"+..
        "?product=Scilab%20software&component=Documentation%20pages"+..
        "&short_desc="+fun
    end
    // URL(known bugs)
    // BugzillaSearch is defined in the calling level
    r.bugs_known = strsubst(BugzillaSearch,"£",fun)
    
    // refname(s)
    e = xmlXPath(doc,"//h1[@class=""refname""]")
    r.refname = xmlAsText(e)'
    
    // refpurpose
    e = xmlXPath(doc,"//p[@class=""refpurpose""]")
    r.refpurpose = uman_cleantxt(xmlAsText(e)', "blanks")
        // Some pages presenting multiple equivalent functions do not repeat
        // the short description for last aliases 
        // (printf.. http://bugzilla.scilab.org/13748)
        // We must complete them here:
    tmp =  r.refpurpose
    if size(tmp,1)<size(r.refname,1) then
        r.refpurpose = [tmp ; emptystr(size(r.refname,1)-size(tmp,1),1)+r.refpurpose($)]
    end
    
    // synopsis title
    CS = struct("en","Calling Sequence",..
                "fr","Usages",..
                "ja","呼出し手順", ..
                "pt","Seqüência de Chamamento",..
                "de","Aufruf", ..
                "ru","Последовательность вызова",..
                "zh","调用序列",..
                "fa","دنباله‌ی فراخوانی")
    if isfield(CS,lang) then
        r.synopsisTitle = CS(lang)
    else
        e = xmlXPath(doc,"//div[@class=""refsynopsisdiv""]")
        if e.size>0 then
            r.synopsisTitle = e(1).children(1).content
        else
            r.synopsisTitle = _("uman", "Calling Sequence")
        end
    end

    // Synopsis
    //s = xmlXPath(doc,"//div[@class=""synopsis""]")
    [s,e,M] = regexp(htmlcat, "%<div class=""synopsis""><pre>(.*?)</pre>%")
    tmp = []
    for m = M'
        s = strsplit(uman_cleantxt(m,"tags|htmlent|style"),"££")
        tmp = [tmp ; s ]
        if size(s,1)>1 then
            tmp = [tmp ; ""]
        end
    end
    if tmp($)=="" then // removing the last blank line
        tmp($) = []
    end
    if size(r.refname,1)==1 then    // Removing blank lines
        tmp(find(tmp=="")) = []
    end
    r.synopsis = tmp
    
    // REFSECTIONS
    // ===========
    refsections = xmlXPath(doc,"//div[@class=""refsection""]")
    nrs = length(refsections)

    // MAIN LOOP over SECTIONS of all TYPES
    // ------------------------------------
    for i = 1:nrs
        rs = refsections(i)
        secTitle = xmlXPath(rs, "./h3[@class=""title""]")
        if size(secTitle,"*")>0
            tmp = uman_cleantxt(secTitle(1).content, "blanks")
            [secType, typeName] = typeof_section(tmp, lang)
            r.sections(i).title = typeName
            
            contents = uman_xml2txt(rs, "")
            contents(1) = []     // suppressing the embedded title
        else
            secType = "description"
            r.sections(i).title = ""
            contents = uman_xml2txt(rs, "")
        end
        r.sections(i).type  = secType
        if secType=="examples" then
            contents = uman_strsubst(contents, "/\[IMAGE\: .*?\]/","","r")
        elseif secType=="see_also"
            k = find(stripblanks(contents)=="")
            contents(k) = []
        end
        while contents(1)=="" | grep(contents(1),"/^\s*$/","r")~=[]
            contents(1) = []
        end
        while contents($)=="" | grep(contents($),"/^\s*$/","r")~=[]
            contents($) = []
        end
        
        r.sections(i).contents = contents
    end

    //  History : de = Historische
    xmlDelete(doc)
endfunction

// ===================================================================

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()

function [secType, typeName] = typeof_section(secTitle, lang)
    // id vs titles for sections    
    langs = [ "en" "de" "fr" "ja" "pt" "ru" "zh" "fa"]
    ID = ["parameters"              "description"  "examples" ]
    en = ["Parameters,Parameter,arguments"    "Description"  "Examples,example"]
    de = ["Parameter"               "Beschreibung" "Beispielen,beispiel,beispiele"]
    fr = ["Paramètres"              "Description"  "Exemples,exemple"]
    ja = ["設定,引数,パラメータ"       "説明"         "例"]
    pt = ["Parâmetros,Argumentos"   "Descrição"    "Exemplos,exemplo"]
    ru = ["Аргументы"               "Описание"     "Примеры,Примерами,ример,примеров"]
    zh = ["設置"                    "描述"          "示例,例子"]
    fa = ["آرگومان‌ها"               "توصیف"        "مثال‌ها"]

    ID = [ ID "see_also"    "history"   ]
    en = [ en "See Also"    "History"   ]
    de = [ de "Siehe auch"  "Historische"]
    fr = [ fr "Voir aussi"  "Historique"]
    ja = [ ja "参照"         "履歴"      ]
    pt = [ pt "Ver também"  "Histórico" ]
    ru = [ ru "Смотрите также" "История"]
    zh = [ zh "参见"         "历史的"    ]
    fa = [ fa "همچنین ببینید" "تاریخچه"  ]  // "See also" in Farsi (technical bug to be fixed)

    k = find(langs==lang)
    if k==[] then
        lang = "en"
    end
    execstr("Ref = "+lang)
    secType = "description" // default
    typeName = secTitle     // default
    L = length(secTitle)
    secTitle= convstr(secTitle)
    
    ref = convstr(Ref)
    found = %f
    while ~found
        for i=1:size(ref, 2)
            tmp  = tokens(ref(i), ",") 
            tmp2 = tokens(Ref(i), ",")
            [j,k] = grep(secTitle, tmp)
            if j~=[] & (i~=1 | or((L-length(tmp(k)))<3)) then
                secType = ID(i)
                if or((L-length(tmp(k)))<3)
                    typeName = tmp2(1)
                end
                found = %t
                break
            end
        end
        tmp = convstr(en)
        if found | and(ref==tmp), break, end
        ref = tmp
    end
endfunction
// ===================================================================

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()
function jarList = uman_listJar()
    // Returns de column matrix of paths to ALL JAR files (not only for LANG or en_US)
    
    function flist = dirjar(path)
        // path must be ended with filesep()
        if isdir(path) then
            if getos()=="Windows" then
                flist = unix_g("dir /s /b "+path+"scilab_*_help.jar 2> nul")
            elseif getos()=="Linux"
                flist = unix_g("find "+path+" -name scilab_*_help.jar 2> /dev/null")
            else
                flist = unix_g("find 2>/dev/null "+path+ "scilab_*_help.jar")
            end
            if flist=="" then
                flist = []
            end
        else
            flist = []
        end
    endfunction
    
    jarList = []
    // Scilab trunk:
    path = pathconvert(pathconvert(SCI)+"modules\helptools\jar\")
    jarList = [jarList ; dirjar(path)]
    // Contrib (ATOMS and others) : ALL USERS
    path = pathconvert(pathconvert(SCI)+"contrib\")
    jarList = [jarList ; dirjar(path)]
    // USER : ATOMS
    path = pathconvert(pathconvert(SCIHOME)+"atoms\")
    jarList = [jarList ; dirjar(path)]
    // USER : other contribs
    path = pathconvert(pathconvert(SCIHOME)+"contrib\")
    jarList = [jarList ; dirjar(path)]
endfunction

// -------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()
function extractJhelpMaps(jarList, atoms, unzip, manDir, mapDir)
    // From the list of selected JAR files, for each file, extracts its 
    //  jhelpmap.jhm in a common directory, and renames it according to the
    //  module
    fs = filesep()
    pattern = pathconvert("\modules\helptools\jar\scilab_", %f)
    for jar = jarList'
        contrib? = %f
        if grep(jar, pattern)~=[]
            module = "scilab"
        else
            if atoms~=[]
                [tmp, ind ] = grep(getshortpathname(jar), atoms(:,4))
            else
                ind = []
            end
            if ind~=[] 
                module = atoms(ind(1),1)    // Name = id in ATOMS
            else
                // The name of the module (such as stored in %helps(:,2))
                //  can be retrieved through the file name
                //  ~\etc\modulename.start
                tmp = pathconvert(dirname(jar)+fs+"..\etc\*.start", %f)
                if getos()=="Windows" then
                     startfile = unix_g("dir /b " + tmp + " 2> nul")
                elseif getos()=="Linux"
                    startfile = unix_g("ls -1 " + tmp + " 2> /dev/null")
                else
                    tmp = pathconvert(dirname(jar)+"\..\etc *.start", %f)
                    startfile = unix_g("find 2>/dev/null "+tmp)
                end
                module = basename(startfile)
                contrib = %t
            end
        end
        lang = uman_strsubst(jar, "|.*?_([a-z]{2}_[A-Z]{2})_help.jar|", "$1", "r")
        subDir = "scilab_"+lang+"_help"+fs
        zippedMap = subDir + "jhelpmap.jhm"
        command = unzip + jar + " " + zippedMap + " -d "+manDir
        status = host(command)
        unzippedMapFrom = manDir + subDir + "jhelpmap.jhm"
        unzippedMapTo   = mapDir + module + ".jhm" 
        [s, m] = movefile(unzippedMapFrom, unzippedMapTo)
    end
endfunction

// -------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()
function out = uman_cleantxt(txt, cleanjobs)
    // cleanjobs: "tags|htmlent|catlines|style|blanks"
    cleanjobs = tokens(cleanjobs,"|")'
    out = txt
    for cleanjob = cleanjobs
        select cleanjob
        case "tags"
             out = uman_strsubst(out, "/<.*?>/", "", "r")
         case "catlines"
             for i = 1:size(out,"*") 
                out(i) = strcat(stripblanks(strsplit(out(i),"££"))," ")
             end
         case "htmlent"
             patterns = ["&#0034;" "&#0038;" "&#0039;" "&#0060;" "&#0062;" "&gt;" "&lt;" "&quot;" ]
             replaces = [ """"       "\&"       "''"      "<"       ">"       ">"   "<"   """"    ]
            out =  uman_strsubst(out, patterns , replaces)
         case "style"
            out =  uman_strsubst(out, "/([^ \=<>~])=([^ \=''""])/", "$1 = $2", "r")
            out =  uman_strsubst(out, "/([^\s]);([^\s])/", "$1; $2", "r")
        case "blanks"
            out = uman_strsubst(out, "/(\s{2,})/", " ", "r")
        end
    end
endfunction

// -------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()

function out = uman_wraplines(text, ncol)
    // LANG must be set in the calling environnement
    if text==[] then
        out = []
        return
    end
    if or(LANG==["ja_JP" "zh_CN"]) then
        nchar = floor(ncol/1.4)  // because symbols are wider than latin ones
    else
        nchar = ncol
    end
    nchar = nchar-1
    text = stripblanks(text)
    Sep = " "
    if LANG=="ja_JP" then
        Sep = [Sep "。" "・"]
//        [T, sep] = strsplit(text, Sep)
//        sep($+1) = ""
//    else
//        if text~=[]
//            T = tokens(text," ")    // 30x faster than strsplit...
//            sep = emptystr(T)+" "
//        else
//            T = []
//            sep = ""
//        end
    end
    [T, sep] = strsplit(text, Sep)
    sep($+1) = ""
    LT = length(T)+1
    out = []
    line = ""   // current line
    lineL = 0   // lineLength
    for i = 1:size(T,"*")
         if lineL+LT(i)<=nchar | line==""
             line = line + T(i) + sep(i)
             lineL = lineL + LT(i)
         else
             if lineL+LT(i)>nchar & LANG=="ja_JP" & grep(line,"/(—\s|[^。・\s])$/","r")~=[]
                 line = line + T(i) + sep(i)
                 out = [out ; line]
                 line = ""
                 lineL = 0
             else
                 out = [out ; line]
                 line = T(i) + sep(i)
                 lineL = LT(i)
             end
         end
    end
    if stripblanks(line,%t)~="" then
        out = [out ; line]
    end
    out = stripblanks(strsubst(out, "¤"," "))
endfunction

// -------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()

function [txt, inline] = uman_xml2txt(node, lineHead)
    // Upstream variables used: options, secType
    //
    // Pieces of HTML code (template to decode):
    //<code..>bla bla</code>   : =>  [bla bla]  // cancelled
    //<a href=""..>bla bla</a> : => _{bla bla}  // cancelled
    // <ul class="itemizedlist">
    //       <li><p class="para">..</p> : => nouvelle ligne à chaque <p..>
    //           <p..>...</p>
    //       </li>
    // </ul>
    // <img ..> among which 
    //   <img src="ScilabNote.png"/> and <img src="ScilabWarning.png"/>
    //   <img src="ScilabTip.png"/> and <img src="ScilabCaution.png"/>

    inline = %F
    thistxt = []
    thisline = []
    name = convstr(node.name)
    nc = node.children.size     // number of children
    cw = lines()
    cw = min([cw(1) 100])   // console width. max width limited to 100 chars
    // ---------------------------------------------------------------
    if name=="code" then
        txt = strsubst(uman_cleantxt(node.content, "style|blanks"), " ", "¤")
        inline = %T
        
    elseif or(name==["text" "a" "i"]) then
        txt = uman_cleantxt(node.content, "blanks")
        inline = %T
        
    // ---------------------------------------------------------------
    elseif or(name==["b" "em"]) then
        txt = uman_cleantxt(node.content,"blanks")
        if  name=="em" then
            txt = "/"+txt+"/"
        else
            txt = convstr(txt,"u")
        end
        inline = %T
        
    // ---------------------------------------------------------------
    elseif name=="varname"
        txt = "[" + uman_cleantxt(node.content, "blanks") + "]"
        inline = %T

    // ---------------------------------------------------------------
    //    elseif name=="a" 
    //        // <a> over complex content will ignore content's format :
    //        //txt = "_{" + node.content + "}"
    //        txt = node.content
    //        inline = %T

    // ---------------------------------------------------------------
    elseif or(name==["h3" "div" "p" "span" "li" "dd"])
        if name=="div" & node.attributes.class~=[] & node.attributes.class=="programlisting" 
            k = find(preTableNoTag==strsubst(node.content," ",""))
            if k~=[]
                txt = [ strsplit(preTable(k(1)),"££") ; ""]
            else
                txt = ""
            end
        else
            // Wrapping width
            // wW = cw-length(lineHead)-3    
            wW = cw - length(lineHead)
            if wW<30, wW = 30, end            // minimal width = 30 chars
            //
            forceInLine = (name=="li" & isdef("simpleContent", "n"))
            if name=="li" & forceInLine
                tmp = xmlXPath(node, "./ul")  // are there any nested sub-list?
                if size(tmp, "*")>0         //   yes:
                    forceInLine = %f
                end
            end

            for i = 1:nc
                [ tmp, inL ] =  uman_xml2txt(node.children(i), lineHead)
                if inL | forceInLine then
                    thisline = thisline + strcat(tmp, " ")
                else 
                    tmp2 = uman_wraplines(uman_cleantxt(thisline,"blanks"), wW)
                    thistxt = [ thistxt ; tmp2 ; tmp ]
                    thisline = []
                end
            end
            if thisline~=[]
                tmp2 = uman_wraplines(uman_cleantxt(thisline,"blanks"), wW)
                txt = [ thistxt ; tmp2 ]
            else
                txt = thistxt
            end
            if name=="span" & size(txt,1)==1 then
                inline = %T
            end
        end
        
    // ---------------------------------------------------------------
    elseif name=="pre"
        // preTable, preTableNoTag = external variables 
        k = find(preTableNoTag==strsubst(node.content, " ", ""))
        txt = strsplit(preTable(k(1)), "££")

    // ---------------------------------------------------------------
    elseif name=="img"  // ex: <img src='jarsci:./PlotSparse_1.png'/>
        // works with Scilab pages. Untested with external packages
        tmp = node.attributes.src;
        tmp = tokens(tmp,"/")
        txt = "[IMAGE: "+tmp($)+"]"
        inline = %t

    // ---------------------------------------------------------------
    elseif name=="dl"     // definitions list
        // <dl> : definitions list : ignorer content ; traiter les children
        //   <dt><span class="term">..</span>      : definition term
        //       <dd><p class="para">...</p></dd>  : definition definition
        //   </dt>
        // </dl>
        // Processing all <dt> entries
        terms = []     // matrix of terms
        tmp = getversion("scilab");  sci6 = evstr(tmp(1))>5;  
        defs  = cell() // cell of related definitions
        i = 1    // n° of <dl>'s children
        it = 0   // n° of terms
        maxL = 8   // maximal length of "alignable" terms
        minL = 5    // minimal indentation of lines for the bloc
        while i<=nc
            // only <dt>..</dt> should be considered:
            while node.children(i).name~="dt" & i<nc
                i = i+1
            end
            dt = node.children(i)
            sp = dt.children(1) // should be <span class="term">..</span>
            it = it+1            
            tmp = stripblanks(sp.content)
            // trimming any ending ":" (as in the ode page)
            L = length(tmp)
            if L>0 & part(tmp,L)==":"
                terms(it)  = part(tmp,1:L-1)
            else
                terms(it) = tmp
            end
            
            // looking for the forthcoming <dd> section:
            sdt = dt.children.size
            j = min([2,sdt])   // n° of <dt>'s children
            while j<sdt & convstr(dt.children(j).name)~="dd" 
                j = j+1
            end
            //dd = dt.children(j) should be <dd><p class="para">...</p></dd>
            // local header length (estimated)
            k = find(length(terms)<maxL)
            if k~=[]
                hL = max(length(terms(k)))+3
            else
                hL = minL
            end
            ddtxt = uman_xml2txt(dt.children(j), lineHead+part(" ",ones(1:hL)))
            if sci6
                execstr("defs{it} = ddtxt")
            else
                execstr("defs(it).entries = ddtxt")
            end
            // nothing is expected between </dd> and the embedding </dt>
            i = i+1
        end

        // Building the whole set of <dt> entries
        // --------------------------------------
        // For short terms, the description is appended to term's line :
        // Setting the standard width of items label field
        I = find(length(terms)<maxL)
        if I~=[] then
            nC = max(length(terms(I))) + 3  // +3 for " : "
            terms(I) = justify(terms(I),"l")
        else
            nC = minL
        end
        // Setting the local blank header
        header = part(" ", ones(1:nC))
        
        // Setting the standard width of item Description field
        wD = max([30, cw - length(lineHead+header) ])
        
        // Loop on items
        for L = 1:size(terms,1)
            wD_line1 = wD - max([ 0 (length(terms(L))+3-nC)])
            if sci6
                execstr("tmp = defs{L}")
            else
                execstr("tmp = defs(L).entries")
            end
            if size(tmp,1)<2 then
                tmp2 = uman_wraplines(tmp, wD_line1+1)
                thistxt =[thistxt; terms(L)+" : "+tmp2(1)]
                if tmp~=tmp2(1) then
                    tmp2 = uman_wraplines(strcat(tmp2(2:$)," "), wD)
                    thistxt  = [thistxt; ""; header + tmp2]
                end
            else
                if wD_line1 < wD
                    // The first line is kept clear
                    thistxt =[thistxt; terms(L)+" : " ;  header+tmp ]
                else
                    thistxt =[ thistxt ; terms(L)+" : "+tmp(1); header + tmp(2:$)]
                end
            end
            if size(tmp,1)>1 & stripblanks(thistxt($))~="" then
                thistxt($+1) = ""
            end
        end
        txt = " "+thistxt
    // ---------------------------------------------------------------
    elseif or(name==["ul" "ol"])
        // Processing all <li> entries :
        litab = []
        for i = 1:nc
            li = node.children(i)
            tmp = uman_xml2txt(li, "  "+lineHead)
            if name=="ul"
                tmp(1) = "* "+tmp(1)
            else
                tmp(1) = sprintf("%2d",i)+". "+tmp(1)
            end
            if size(tmp,1)>2 & stripblanks(tmp($))~="" then
                tmp($+1) = ""
            end
            tmp(2:$) = "  "+tmp(2:$)
            litab = [ litab ; tmp ]
        end
        if size(litab,1)>8 then
            litab = [""; litab]
        end
        txt = " "+litab
        
        // Post-processing for the list of summaries, in order to align the "--"
        if grep(options, "s")~=[] | secType=="see_also"
            K = grep(txt, "/^ \*.*?—/","r")
            Kw = grep(txt, "/^   [^\s\*]/", "r")  // wrapped line
            nL = size(txt,1)
            if K~=[] then
                txt1 = stripblanks(txt(K))
                s = strcspn(txt1,"—")            // locating the "—" separator
                ms = mean(s)
                ss = stdev(s)
                p = min([18, round(ms+ss)])
                if ss<7 then
                    p = min([18, round(ms+1.5*ss)])
                end
                j = find(s>0 & s<p)       // # of lignable lines
                t = []
                for m = j
                    t = [ t ; strsplit(txt1(m),s(m))']
                end
                p = min([18, max(length(t(:,1)))+1])
                t = justify(t(:,1), "l") + " " + t(:,2)
                txt1(j) = t
                txt(K) = txt1
                // Shifting wrapped lines
                head = part(" ", ones(1,p+3))
                ww = max([30, cw - length(lineHead)-p]) // wrapping width
                
                newtxt = []
                line = ""
                for i = 1:size(txt,"*")
                    if or(i==Kw) 
                        line = line + " "+ stripblanks(txt(i))
                    else
                        if line~=""
                            // we process the registered line
                            tmp = uman_wraplines(line, ww+p+3)
                            newtxt = [ newtxt ; " "+tmp(1) ]
                            if size(tmp,1)>1
                                tmp = uman_wraplines(strcat(tmp(2:$), " "), ww)
                                newtxt = [newtxt ; head + tmp ]
                            end
                            line = " "
                        end
                        if or(i==K)
                            line = " "+txt(i)
                        else
                            newtxt = [newtxt ; txt(i)]
                        end
                    end
                end
                if line~=" "
                    // we process the registered line
                    tmp = uman_wraplines(line, ww+p+3)
                    newtxt = [ newtxt ; " "+tmp(1) ]
                    if size(tmp,1)>1
                        tmp = uman_wraplines(strcat(tmp(2:$), " "), ww)
                        newtxt = [newtxt ; head + tmp ]
                    end
                end
            else
                newtxt = txt
            end
            txt = newtxt
        end
//     ---------------------------------------------------------------
    elseif or(name==["table" "tbody"])
        // <note> and <warning> are in a table starting with an image:
        //   <table><tr><td valign="top"><img src="ScilabNote.png"/>
        tmp = node.children(1).children(1).children(1)
        if tmp.name=="img"
            if grep(tmp.attributes.src, ["ScilabNote.png" "ScilabTip.png"])~=[]
                txt = _("uman", "NOTE: ")
            elseif grep(tmp.attributes.src, ["ScilabWarning.png" "ScilabCaution.png"])~=[]
                txt = _("uman", "WARNING: ")
            else
                txt = ""
            end
            txt = txt + uman_cleantxt(node.content+" ","blanks")
            txt = uman_wraplines(txt, max(30, cw-length(lineHead)))
            return
        end
        a = node.attributes
        TABLE = %t
        Array = []
        for i = 1:nc        // Loop over rows
            row = uman_xml2txt(node.children(i)) 
            if i>1 
                // missing cells?
                if size(row,2)< size(Array,2)
                    row = resize_matrix(row, 1, size(Array,2))
                elseif size(row,2)> size(Array,2)
                    Array = resize_matrix(Array, 1, size(row,2))
                end
            end
            Array = [Array ; row ]
        end
        // If there are empty (or blank) columns => we remove them
        L = length(stripblanks(strcat(Array,"","r")))
        Array(:, find(L==0)) = []
        
        Array = justify(Array, "l")
        if a.border~=[] & a.border~="0"
            Array = strcat(Array,"| ","c")
            Array = "| " + Array + " |"
            hline = part("-",ones(1,length(Array(1,:))))
            txt = " " + [hline ; Array ; hline ]
        else
            Array = strcat(Array," ","c")
            if size(Array,1)==1 then
                txt = stripblanks(Array)
            else
                txt = " " + Array
            end
            
        end
    // ---------------------------------------------------------------
    elseif name=="tr"
        Row = node.children
        row = []
        if nc>1
            // We assume that <td> cells have simple contents:
            for i = 1:nc
                tmp = uman_cleantxt(Row(i).content+" ","blanks")
                if Row(i).name=="th"
                    tmp = convstr(tmp,"u")
                end
                row($+1) = tmp
            end
            txt = row.'
        else
            txt = uman_xml2txt(Row(1), lineHead)
        end
    // ---------------------------------------------------------------
    else 
        txt = "<"+name+">"+node.content+"</"+name+">"
    end
endfunction

// ------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()
function uman_openURL(url, msg)
    if getos()=="Windows"
        winopen(url)
    elseif getos()=="Linux"
        unix("xdg-open """+url+"""")
    elseif getos()=="Darwin"         // Mac OS
        unix("open """+url+"""")
    else
        warning(msg)
    end
endfunction

// ------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()
function rep = redirect_from_external(item, options)
// List of equivalences or nearest redirections
//  column #1: external item (Octave, etc...)
//  column #2: Scilab redirection (help's filename).
//             "£" prefix = options to be passed when calling uman
e = [ "all"     "and"
      "angle"   "atan"
      "any"     "or"
      "arrayfun" "feval"
      "atan2"   "atan"
      "axes"    "newaxes"
      "axis"    "gca"
      "balance" "balanc"
      "box"     "gca"
      "caxis"   "colorbar"
      "chebwin" "window"
      "cla"     "delete"
      "clear global"    "clearglobal"
      "closereq" "xdel"
      "colormap" "jetcolormap"
      "colors"   "d£color_list"       // inner redirection
      "compan"   "companion"
      "contourc" "contour2di"
      "contour3" "contour"
      "copyobj" "copy"
      "cputime" "timer"
      "csvread" "csvRead"
      "datatypes" "d£type"           // inner redirection with "d" forced
      "dbclear" "delbpt"
      "dbcont"  "resume"
      "dbquit"  "abort"
      "dbstack" "where"
      "dbstatus" "dispbpt"
      "dbstop"  "setbpt"
      "dbup"    "return"
      "deblank" "stripblanks"
      "demo"    "demo_gui"
      "dlmwrite" "fprintfMat"
      "echo"    "mode"
      "eig"     "eigs"
      "ellipj"  "amell"
      "ellipke" "percentk"
      "eps"     "percenteps"
      "errorbar" "errbar"
      "errordlg" "messagebox"
      "false"   "%f"
      "fclose"  "mclose"
      "feof"    "meof"
      "files"   "£listfiles"     // inner summary
      "fftn"    "fft"
      "fgetl"   "mgetl"
      "fgets"   "mgetl"
      "fill"    "xfpolys"
      "flipud"  "flipdim"
      "fliplr"  "flipdim"
      "fopen"   "mopen"
      "fread"   "mget"
      "frewind" "mseek"
      "fwrite"  "mput"
      "fzero"   "fsolve"
      "gallery" "testmatrix"
      "gammainc" "cdfgam"
      "ginput"  "locate"
      "grid"    "xgrid"
      "gtext"   "legends"
      "hdf5"    "£h5read"       // inner summary http://www.hdfgroup.org/HDF5/
      "help"    "uman"          // Octave
      "helpdlg" "messagebox"
      "helpwin" "help"
      "hggroup" "glue"
      "hgload"  "load"
      "hgsave"  "save"
      "hist"    "histplot"
      "histcounts" "histc"
      "history" "£gethistoryfile"   // inner summary
      "hold"    "axes_properties"
      "housh"   "householder"   // Octave
      "horzcat" "cat"
      "i"       "%i"
      "ifft2"   "ifft"
      "ifftn"   "ifft"
      "imagesc" "Matplot1"
      "inf"     "%inf"
      "invfeqz" "frfit"
      "isgraphics" "is_handle_valid"    // Octave
      "ishghandle" "is_handle_valid"    // Octave
      "iskeyword" "getscilabkeywords"
      "ismac"   "getos"
      "ismember" "members"
      "ispc"    "getos"
      "isunix"  "getos"
      "j"       "%i"
      "kaiser"  "window"
      "keyboard" "pause"
      "latex"   "d£math_rendering_features_in_graphic" // inner redirection
      "lenght"  "length"        // common mis-spelling
      "listdlg" "x_choose"
      "loglog"  "plot2d"
      "lookfor" "apropos"
      "lower"   "convstr"
      "magic"   "testmatrix"
      "matlabroot" "SCI"
      "metanet" "£bandwr"       // inner redirection: "uman metanet s" will be OK
      "mfilename"  "get_absolute_file_path"
      "mod"     "pmodulo"
      "more"    "lines"
      "nan"     "%nan"
      "nanmin"  "min"
      "nanmax"  "max"
      "nargin"  "argn"
      "nargout" "argn"
      "nonzeros" "spget"
      "num2str" "string"
      "ode23"   "ode"
      "ode45"   "ode"
      "ode113"  "ode"
      "ode15s"  "ode"
      "odeget"  "odeoptions"
      "paddarray"   "resize_matrix"
      "pi"      "%pi"
      "planerot" "givens"
      "plot3"   "param3d"
      "plottools" "ged"
      "polar"   "polarplot"
      "polyder" "derivat"
      "polyint" "integratePoly"
      "polyval" "horner"
      "printdlg" "toprint"
      "psi"     "dlgamma"
      "questdlg" "messagebox"
      "randi"   "grand"
      "randn"   "rand"
      "randperm" "grand"
      "range"   "strange"
      "rbbox"   "rubberbox"
      "realmax" "number_properties"
      "realmin" "number_properties"
      "regexprep"  "strsubst"
      "rem"     "modulo"
      "reshape" "matrix"
      "rmfield" "null"
      "rot90"   "flipdim"
      "rotate3d" "rotate_axes"
      "run"      "exec"
      "semilogx" "plot2d"
      "semilogy" "plot2d"
      "shg"      "show_window"
      "shortcuts" "d£console"     // inner redirection
      "signal"   "£detrend"     // inner summary
      "sort"     "gsort"
      "sortrows" "gsort"
      "sparses"  "£chsolve"     // inner summary
      "sprandn"  "sprand"
      "sprandsym" "sprand"
      "spy"      "PlotSparse"
      "stairs"   "plot2d2"
      "statistics" "£reglin"    // inner summary
      "std"      "stdev"
      "stem"     "plot2d3"
      "strfind"  "strindex"
      "strjoin"  "strcat"
      "strjust"  "justify"
      "strtrim"  "stripblanks"
      "switch"   "select"
      "tempdir"  "TMPDIR"
      "texlabel" "math_rendering_features_in_graphic"
      "text"     "xstring"
      "textscan" "msprintf"
      "trapz"    "inttrap"
      "trigo"    "£cos"
      "true"    "%t"
      "tukeywin" "window"
      "uisetcolor" "uigetcolor"
      "unix"    "unix_g"
      "upper"   "convstr"
      "var"     "variance"
      "variables" "£who"            // inner summary
      "version" "getversion"
      "vertcat" "cat"
      "view"    "axes_properties"
      "waitforbuttonpress" "xclick"
      "warndlg" "messagebox"
      "which"   "get_function_path"
      "windows"  "£winopen"          // inner summary
      "whos -file" "listvarinfile"
      "xml"     "£xmlRead"          // inner summary
      "xlim"    "axes_properties"
      "ylim"    "axes_properties"
      "zlim"    "axes_properties"
      "zoom"    "zoom_rect"        // + unzoom
    ]
    // REDIRECTIONS to modules, FileExchanges, or external references
    e = [e
      "circshift"   "cshift"
      "inpolygon"   "point_in_polygon"
      "python"      "PIMS"
      "MySQL"       "scidb"
      "MSSQL"       "scidb"
      "Sqlite"      "scidb"
      "PostgreSQL"  "scidb"
      "ODBC"        "scidb"
      "Firebird"    "scidb"
      ]
    
    // Queries matching a scilab item but with a different meaning in an external 
    //  language. There are considered only if the priority target is external package
    if grep(options, "x")~=[] then
        e = [ e
      "contour"  "contour2d"
      "conv"    "convol"
      "dos"     "unig_g"
      "end"     "dollar"
      "home"    "d£tohome"
      "load"    "loadmatfile"
      "null"    "kernel"
      "pause"   "halt"
      "quiver"  "champ"         // quiver implemented in plotlib
      "system"   "unix_g"
      "type"    "unix_w"
      "ver"     "getversion"
      ]
    end
    item = item(1)
    if convstr(item)==item 
        ie = find(convstr(e(:,1))==item)
    else
        ie = find(e(:,1)==item)
    end
    if ie~=[]
        rep = e(ie(1),2)
    else
        rep = item
    end

endfunction

// -------------------------------------------------------------------------

// Copyright (C) - 2015 - Samuel GOUGEON
// Utility function for uman()
function table = get_atoms_forge()
    // Actually, the table may also contain references of non-ATOMS packages
    // [ module_name  module_name_on_forge  website ]
    atoms = [
    "accsum"            "accsum"    ""
    "AdaptiveOptics"    ""          "http://sciao.sourceforge.net/#authors"
    "ADS_CoLiSyS"       ""          ""
    "aerospace-toolbox" "aerospace-toolbox"	""
    "aerospace_blockset" "aerospace-blockset"	""
    "ampl_toolbox"      ""          "http://code.google.com/p/scilab-mip"
    "Analog_and_DigitalCommunication"   ""  "http://www.geocities.ws/senthilirtt/"
    "Anfislab"          ""          "http://www.tritytech.com"
    "ANN_Toolbox"       "ann-toolbox" ""
    "apifun"            "apifun"    ""
    "ardronemodule"     ""          "https://code.google.com/p/ardronetoolbox/"
    "arduino"           "arduino"   ""
    "arfit"             ""          "http://www.gps.caltech.edu/~tapio/arfit/"
    "assert"            "assert"    ""
    "BackDoor"          "BackDoor"  ""
    "BTD"               ""          "http://dcb.res.in/scilab-btd/"
    "casci"             "casci"     "http://philippe.castagliola.free.fr"
    "CCA"               ""          "http://people.ok.ubc.ca/ylucet/cca.html"
    "celestlab"         "celestlab" ""
    "celestlabx"        ""          ""
    "cglab"             "cglab"     ""
    "CLUSTER"           ""          "http://www.xing.com/profile/Harald_Galda"
    "condnb"            "condnb"    ""
    "Conint"            "conint"    ""
    "conmin"            ""          "http://code.google.com/p/scilab-mip"
    "convertlatin"      "convertlatin"  ""
    "coselica"          "coselica"  ""
    "CMA-ES"            ""          "http://www.cfuttrup.com"
    "CPGE"              "cpge"      ""
    "csv_readwrite"     "csv-readwrite" ""
    "Curl"              ""          ""
    "CUTEr"             "cuter"     ""
    "CWA"               "CWA"       ""
    "dace_scilab"       "dace-scilab"   ""
    "dbldbl"            "dbldbl"    ""
    "dde_toolbox"       "dde-toolbox"   ""
    "DD_QD"             ""          "http://www.mi.kagu.tus.ac.jp/qupat.html"
    "DeCaToKi"          ""          "http://sites.google.com/site/ronaystein/"
    "diffcode"          "diffcode"  ""
    "digimetrie_Tools"  ""          "http://www.digimetrie.com/"
    "dispmat"           "dispmat"   ""
    "distfun"           "distfun"   ""
    "Dynpeak"           ""          ""
    "edflib"            ""          "http://www.teuniz.net/edflib/"
    "emd_toolbox"       ""          "http://perso.ens-lyon.fr/patrick.flandrin/emd.html"
    "EMTTOOL"           ""          ""
    "FACT"              ""          ""
    "femtruss"          ""          "http://code.google.com/p/scilab-mip/"
    "Financial"         "Financial" ""
    "fitters"           ""          ""
    "floatingpoint"     "floatingpoint" ""
    "fmincon"           "fmincont"  ""
    "fsmtbx"            ""          "http://fsmtbx.origo.ethz.ch/"
    "getDmatrix"        ""          "http://uolp.unifor.br/oul/pages/academico/posgraduacao/novoSite/professorPL.jsp?Professor=3820065&p_tipo_pagina=grad"
    "Global_Optim_toolbox"  ""      "http://www.sc.iitb.ac.in/~bhagyesh/Scilab/GlobalOptimizationToolbox.html"
    "grocer"            ""          "http://dubois.ensae.net/grocer.html"
    "guibuilder"        ""          "http://www.tritytech.com/"
    "guimaker"          "guimaker"  ""
    "helptbx"           "helptbx"   ""
    "huffcomp"          ""          ""
    "hypt"              "hypt"      ""
    "hyb_auto_module"   "hyb-auto-module"   ""
    "HYDROGRv50"        ""          "http://lerat.julien.free.fr/CONT_FR/HYDROGR_FR.html"
    "ica_toolbox"       ""          "http://research.ics.tkk.fi/ica/fastica/"
    "identification"    ""          ""
    "imsls"             "imsls"     ""
    "indinf"            "indinf"    ""
    "intprbs"           "intprbs"   ""
    "iodelay"           ""          ""
    "IPD"               "IPD"       "http://www.xing.com/profile/Harald_Galda/"
    "ipoptlib"          "ipoptlib"  ""
    "IPT2"              ""          "http://cdac.in"
    "IPT3"              ""          ""
    //"IWT"               ""          "http://www.cdac.in": replaced with the next
    "Integer_Wavelet_Transform_ToolBox" ""  ""
    "JIMS"              "JIMS"      ""
    "json"              "json"      ""
    "krisp"             ""          ""
    "lcc_windows"       "lcc-windows"   ""
    "leftwavedenoise"   ""          "http://cas.ensmp.fr/~chaplais/fr/Ondelettes/Ondelettes/DebruitageOndelettes.html"
    "libsvm"            "libsvm"    ""
    "linalg"            "linalg"    ""
    "lolimot"           "lolimot"   "http://code.google.com/p/scilab-mip"
    "lowdisc"           "lowdisc"   ""
    "lsitbx"            ""          "http://lsitbx.origo.ethz.ch/"
    "makematrix"        "makematrix"    ""
    "maple2scilab"      "maple2scilab"  ""
    "Mascaret"          ""          ""
    "Mathieu"           "mathieu"   ""
    "MatrixMarket"      "matmark"   ""
    "Maxplus"           ""          ""
    "maxplus_petrinet"  ""          ""
    "Max_Plus_Algebra"  ""          "http://www.its.ac.id/personal/index.php?id=Subiono"
    "MDPtoolbox"        ""          "http://www.inra.fr/mia/T/MDPtoolbox"
    "MED"               "MED"       ""
    "metanet"           "metanet"   ""
    "mi"                ""  "http://penglab.janelia.org/proj/mRMR/"
    "microdaq"          ""          ""
    "microwave"         ""          "http://www.reveyrand.fr/uW.html"
    "mingw"             "MinGW-toolbox" ""
    "minphase_toolbox"  ""          ""
    "mkltools"          "mkltools"  ""
    "moc_toolbox"       "moc-toolbox" ""
    "modbus"            "modbus"    ""
    "module_lycee"      "module-lycee"  ""
    "montesci"          ""          "http://www.drivehq.com/web/ka_eins/montesci/"
    "MPScilab"          "mpscilab"  ""
    "nan"               "nan-toolbox"   ""
    "NARVAL"            ""          "http://wwwfr.uni.lu/snt/"
    "NIDAQ"             "nidaq"     ""
    "NISP"              "nisp"      ""
    "nistdataset"       "nistdataset"   ""
    "Nqueens"           ""          ""
    "NTG"               ""          "http://wwwfr.uni.lu/interdisciplinary_centre_for_security_reliability_and_trust/people/foued_melakessou"
    "number"            "number"    ""
    "OFIP"              ""          ""
    "opc_client"        "opc-client" ""
    "OpenRTDynamics"    ""          ""
    "opti"              ""          ""
    "optkelley"         "optkelley" ""
    "Ortpol"            "ortpol"    ""
    "Overload_Toolbox"  "overload"  ""
    "PIMS"              "pims"      ""
    "plotlib"           "plotlib"   ""
    "portaudio"         "portaudio" ""
    "predictivemodule"  ""          ""
    "ProActiveConnector" ""         "http://proactive.activeeon.com"
    "PSO"               "pso-toolbox" ""
    "PVM"               "pvm"       ""
    "quapro"            "quapro"    ""
    "rdataset"          "rdataset"  ""
    "regtools"          "regtools"  ""
    "rest_client"       ""          ""
    "RLDesign"          ""          ""
    "rltool"            ""          ""
    "Robotics_Toolbox"  ""          ""
    "scetoexe"          "scetoexe"  ""
    "Scholar_Lycee_in_English"  "module-lycee"  ""
    "scibench"          "scibench"  ""
    "scicheckupdate"    "scicheckupdate"    ""
    "scicobyla"         "scicobyla" ""
    "SciCrypt"          "scicrypt"  ""
    "SciCSDP"           ""          "http://code.google.com/p/scilab-mip/"
    "scidb"             "scidb"     ""
    "scidemo"           "scidemo"   ""
    "scidoe"            "scidoe"    ""
    "sciFLT"            ""          "http://www.geocities.ws/jaime_urzua/sciFLT/sciflt.html"
    "SciFreeFEM"        ""          "http://code.google.com/p/scilab-mip/"
    "SciGLPK"           ""          "http://code.google.com/p/scilab-mip/"
    "sciGPGPU"          ""          ""
    "scilab2c"          "scilab2c"  ""
    "Scilab_SSE"        ""          "http://www.scilab.org"
    "Scilab_XLL"        "scilab-xll"    ""
    "scilib_geo"        ""          "http://scilib-geo.area7.de/"
    "SciLPSolve"        ""          "http://code.google.com/p/scilab-mip/"
    "scinlopt"          ""          "http://code.google.com/p/scilab-mip"
    "sciopustc"         ""          "http://opus-project.fr/"
    "scipad"            ""          "http://sourceforge.net/projects/scipad/"
    "sci_ipopt"         "sci-ipopt" ""
    "Scholar_Lycee_in_English"  ""  "http://scilab-scholar.blogspot.com"
    "serial"            ""          "github.com/sengupta/Scilab-Serial"
    "serialport"        ""          ""
    "SIMM"              "SIMM"      ""
    "simplex"           ""          "http://code.google.com/p/scilab-mip"
    "SIP"               "SIP"       "http://siptoolbox.sourceforge.net"
    "siseli"            ""          "http://www.askrprojects.net/software/siseli/index.html"
    "SIVP"              ""          "http://sivp.sourceforge.net/"
    "SMP"               "smp"       ""
    "sndfile_toolbox"   "sndfile-toolbox"   ""
    "socket-server"     ""          "https://github.com/ishit/scilab-network"
    "socket_toolbox"    ""          "http://www.reveyrand.fr/SOCKET2.html"
    "specfun"           "specfun"   ""
    "toolbox_spgl1"     "toolbox-spgl1"  ""
    "spilu"             "spilu"     ""
    "splspc"            "splspc"    ""
    "SSP_Toolbox"       ""          "www.cdackolkata.in"
    "Statpack"          "Statpack"  ""
    "stftb"             "stftb"     ""
    "stixbox"           "stixbox"   ""
    "sudoku"            "sudoku"    ""
    "supportvectormachine"  ""      ""
    "swt"               "swt"       ""
    "texmacs"           "texmacs"   ""
    "Two-Dimensional_Transformations"   ""  "www.cdackolkata.in"
    "uman"              ""          "http://sgougeon.free.fr/scilab/help/uman/scilab_en_US_help/uman.html"
    "uncprb"            "uncprb"    ""
    "usb_toolbox"       "usb-toolbox"   ""
    "Wavelab"           ""          "http://www-stat.stanford.edu/~wavelab/Wavelab_850/index_wavelab850.html"
    "WriteXmlExcel"     "writexmlexcel" ""
    "XCPL"              ""          ""
    "xls_link"          "xls-link"  ""
    "xmllab"            "xmllab"    "http://xmllab.org"
    "xmltodocbook"      "xmltodocbook"  ""
    ]
    
    FEroot = "http://fileexchange.scilab.org/toolboxes/"  // list on 2015-05-09
    fileExchange = [
    "addblanks" ""  FEroot+"128000/#comments_remove"
    "bezier"    ""  FEroot+"293000/#comments_remove"
    "BVPsolver" ""  FEroot+"217000/#comments_remove"
    "circfit"   ""  FEroot+"194000/#comments_remove"
    "combine"   ""  FEroot+"237000/#comments_remove"
    "contour3d" ""  FEroot+"contour3d/#comments_remove"
    "cshift"    ""  FEroot+"161000/#comments_remove"
    "cummax"    ""  FEroot+"348000/#comments_remove"
    "cummin"    ""  FEroot+"348000/#comments_remove"
    "dither"    ""  FEroot+"292000/#comments_remove"
    "dxf2sci"   ""  FEroot+"131000/#comments_remove"
    "ellipfit"  ""  FEroot+"194000/#comments_remove"
    "ellipsoid" ""  FEroot+"123000/#comments_remove"
    "ex355p"    ""  FEroot+"341000/#comments_remove"
    "fct"       ""  FEroot+"213000/#comments_remove"
    "FG_ISG500" ""  FEroot+"225000/#comments_remove"
    "gini"      ""  FEroot+"306000/#comments_remove"
    "git"       ""  FEroot+"115000/#comments_remove"
    "ifct"      ""  FEroot+"214000/#comments_remove"
    "integratePoly" "" FEroot+"228000/#comments_remove"
    "isDocked"  ""  FEroot+"360000/#comments_remove"
    "LDS1200"   ""  FEroot+"226000/#comments_remove"
    "legends_mc" ""  FEroot+"274000/#comments_remove"
    "linfit"    ""  FEroot+"250000/#comments_remove"
    "logistics_function" "" FEroot+"124000/#comments_remove"
    "lorenzc"   ""  FEroot+"307000/#comments_remove"
    "mlbs"      ""  FEroot+"mlbs/#comments_remove"
    "mm3136A"   ""  FEroot+"227000/#comments_remove"
    "m2tabular" ""  FEroot+"297000/#comments_remove"
    "M38XR"     ""  FEroot+"232000/#comments_remove"
    "OLPpower"  ""  FEroot+"223000/#comments_remove"
    "orderfields"  ""  FEroot+"349000/#comments_remove"
    "osciOX80"  ""  FEroot+"234000/#comments_remove"
    "point_in_polygon" "" FEroot+"121000/#comments_remove"
    "polyfit"   ""  FEroot+"249000/#comments_remove"
    "polyroots"   ""  FEroot+"362000/#comments_remove"
    "rungekutta" "" FEroot+"151000/#comments_remove"
    "strrepeat" ""  FEroot+"365000/#comments_remove"
    "struct2json" ""  FEroot+"332000/#comments_remove"
    "thermaSnap" "" FEroot+"230000/#comments_remove"
    "triangu3D" ""  FEroot+"291000/#comments_remove"
    "varspace"  ""  FEroot+"364000/#comments_remove"
    "wrscilib"  ""  FEroot+"wrscilib/#comments_remove"
    "XCPL_toolbox" "" FEroot+"/#comments_remove"
    "zaber"     ""  FEroot+"255000/#comments_remove"
    
    "polar2D" ""    FEroot+"285000/#comments_remove"
    "polar3D" ""    FEroot+"285000/#comments_remove"
    "triax" ""      FEroot+"285000/#comments_remove"
    
    "tachometer" ""  FEroot+"312000/#comments_remove"
    "thermometer" "" FEroot+"312000/#comments_remove"
    "variometer" "" FEroot+"312000/#comments_remove"
    "windrose"  ""  FEroot+"312000/#comments_remove"
    
    //"" ""  FEroot+"/#comments_remove"
    //"px2rc" ""  FEroot+"335000/#comments_remove"
    //"strsubst" ""  FEroot+"294000/#comments_remove"
    //"rana" ""           FEroot+"172000/#comments_remove"
    //"thermocouples" ""  FEroot+"248000/#comments_remove"
    //"isprime"     ""    FEroot+"235000/#comments_remove"
    //"peak_detect" ""    FEroot+"209000/#comments_remove"
    //"linear"      ""    FEroot+"118000/#comments_remove"
    ]
    others = [
    "GPIB"      ""  "http://www.reveyrand.fr/GPIB.html"
    "int4sci"   ""  "https://github.com/Int4Sci/Int4Sci-v2"
    "Labview"   ""  "http://www.scilab.org/scilab/interoperability/connection_applications/labview"
    ];
    table = [atoms ; fileExchange ; others]
endfunction
// ==================================================================

// Copyright (C) - 2015 - Samuel GOUGEON
//
// This file must be used under the terms of the CeCILL.
// Its terms are available at
// http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
// 
// Utility function for uman()
function resu = uman_strsubst(text, pattern, replace, option)
    // option: "r" if the pattern is a regular expression (with delimiters and options)
    // replace: Group id: with TCL version = "\#"    with Java = "$#"
    Sclass = "java.lang.String"

    rhs = argn(2)

    if rhs>2 then
        if text==[] then
            resu = []
            return
        end
        
        sizeText = size(text)

        // Chosing the Java method: litteral, regexp, regexp once
        // ------------------------------------------------------
        if rhs>3 then
            if option=="r" then
                jmeth = 1  // "replaceAll"
            elseif or(option==["ro" "or"])
                jmeth = 3  // "replaceFirst"
            else
                msg = _("uman","%s: Wrong value for input argument #%d: Must be in the set {%s}.\n")
                error(msprintf(msg, "strsubst", 4, "''r'',''ro''"))
            end
        else
            jmeth = 2  // "replace"
            option = ""
        end
        
        // Compiling the meta-method
        // -------------------------
        try
            arrayRep = jnewInstance("arrayRep")
        catch
            jimport(Sclass)
            arrayRep = jcompile("arrayRep", ..
            ["public class arrayRep {"
             "public static String[] replaceInArray(int meth, String arr[], String[] pat, String[] rep) {"
             "    int nbelem = arr.length; "
             "    String tmp;"
             "    String[] out = new String[nbelem];"
             "    if (meth==1) { "
             "    for (int i = 0; i < nbelem; i++) { "
             "        tmp = arr[i]; "
             "        for (int p = 0; p < pat.length; p++) { "
             "            tmp = tmp.replaceAll(pat[p], rep[p]); "
             "            } "
             "        out[i] = tmp; "
             "        } "
             "    } "
             "    else if (meth==2) {"
             "    for (int i = 0; i < nbelem; i++) { "
             "        tmp = arr[i]; "
             "        for (int p = 0; p < pat.length; p++) { "
             "            tmp = tmp.replace(pat[p], rep[p]); "
             "            } "
             "        out[i] = tmp; "
             "        } "
             "    } "
             "    else {"
             "    for (int i = 0; i < nbelem; i++) { "
             "        tmp = arr[i]; "
             "        for (int p = 0; p < pat.length; p++) { "
             "            tmp = tmp.replaceFirst(pat[p], rep[p]); "
             "            } "
             "        out[i] = tmp; "
             "        } "
             "    } "
             "    return out; "
             "    }"
             "}"]);
         end
        //pattern = stripblanks(pattern)

        // Pre-processing the patterns: Java does not accept "/../modifiers" syntax.
        //  - modifiers must be inserted as (?#) switches in head of regexp
        //  - delimiters are not expected
        if option~="" then
            // http://docs.oracle.com/javase/tutorial/essential/regex/pattern.html
            // Extracting and checking delimiters:
            delims = part(pattern, 1)
            // Is at least one of opening delimiters a forbiden character?
            forbid = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
            if ~and(strcspn(delims, forbid)==length(delims)) then
                msg = _("uman","%s: Delimiters must be non alphanumeric")
                error(msprintf(msg, "strsubst"))
            end
            // Checking the closing delimiter and extracting modifiers:
            modifiers = strrchr(pattern, delims) // should include the closing delimiter
            Nunclosed = length(find((modifiers==pattern)))
            if Nunclosed>0
                msg = _("uman", "%s: the delimiter in the pattern is sometimes unclosed (%d pattern(s) concerned)")
                error(msprintf(msg, "strsubst", Nunclosed))
            end
            // Supported modifiers:
            // u => (?u) : unicode case: set by default
            // i => (?i) : case insensitive
            // m => (?m) : multiline
            // x => (?x) : ignores comments starting with # until the end of the line
            // s => (?s) : "." matches also EOL
            Modifiers = emptystr(modifiers)+"(?u"
            if or(length(modifiers)>1)
                for m = ["i" "m" "x" "s"]
                    im = grep(modifiers,m)
                    Modifiers(im) = Modifiers(im)+m
                end
            end
            Modifiers = Modifiers + ")"
            pattern = strncpy(pattern, length(pattern)-strcspn(strrev(pattern), delims)-1)
            L = length(pattern)
            for i = 1:length(L)
                pattern(i) = Modifiers(i) + part(pattern(i), 2:L(i))
            end
        end

         // Formating argins
         // ----------------
        text    = text(:)'
        pattern = pattern(:)'
        replace = replace(:)'
        if  size(pattern,"*")==1 then
            replace = replace(1)
        elseif size(replace,"*")==1
            replace = emptystr(pattern) + replace
        elseif size(replace,"*")~=size(pattern,"*")
            msg = _("uman", "%s: Incompatible input arguments #%d and #%d: Same sizes expected.\n")
            error(msprintf(msg, "strsubst", 2, 3))
        end
        
        // Calling the routine
        // -------------------
        resu = arrayRep.replaceInArray(jmeth, text, pattern, replace)
        
        // Post-formatting the result
        resu = matrix(resu, sizeText)

    else
        resu = ""
    end
endfunction
// ------------------------------------------------------------------
// ==================================================================
// CONSTRAINING EXTERNAL BUGS :
//  html parsing : http://bugzilla.scilab.org/12748 (worked around) 
//  parsing "<"  : http://bugzilla.scilab.org/13877 (worked around)
//  parsing <pre>: http://bugzilla.scilab.org/show_bug.cgi?id=12588 (worked around)
//  strsubst     : http://bugzilla.scilab.org/12808  (worked around)
//  diary        : http://bugzilla.scilab.org/show_bug.cgi?id=12551 (ignored)
